from __future__ import absolute_import

import tensorflow as tf
import numpy as np

from skopos.network.simple_layers import FullyConnected
from skopos.utils.preprocessing import Preprocessing
from skopos.techniques.learners import PolicyGradientLearner

class AdvantageActorCritic(PolicyGradientLearner):

	def __init__(self, value_reg=0.5, entropy_reg=0.01):
		super(AdvantageActorCritic, self).__init__()
		self.value_reg = value_reg
		self.entropy_reg = entropy_reg

	def define_placeholders(self):
		x = tf.placeholder(tf.float32, shape=[None, self.network.get_state_dimension()])
		y_ = tf.placeholder(tf.float32, shape=[None])
		actions = tf.placeholder(shape=[None], dtype=tf.int32)
		return x, y_, actions

	def output(self, x, reuse):
		policy_layer = FullyConnected(size=self.network.get_action_number(), initializer=self.columns_initializer(0.01))
		with tf.variable_scope(self.network.get_scope() + 'policy_layer', reuse=reuse):
			policy = policy_layer.apply_layer(x)
			policy = tf.nn.softmax(policy)

		value_layer = FullyConnected(size=1)
		with tf.variable_scope(self.network.get_scope() + 'value_layer', reuse=reuse, initializer=self.columns_initializer(1.0)):
			self.value = value_layer.apply_layer(x)
		return policy

	def prediction(self, out):
		out = out.flatten()
		probabilities = out.clip(min=0)
		probabilities /= probabilities.sum()
		action = np.random.choice(out, p=probabilities)
		action = np.argmax(out == action)
		return action

	def error_function(self, y, y_, actions):
		""" Renaming the variables to better understand """
		policy = y
		value = self.value
		discounted_rewards = y_

		actions_ohe = tf.one_hot(actions, self.network.get_action_number(), dtype=tf.float32)
		
		##### Policy loss ###### 
		""" We add a small constant to prevent a NaN error that could occur 
		if we selected an action while it is probability was zero. """
		logp = tf.log(tf.reduce_sum(policy * actions_ohe, [1]) + 1e-10)
		advantages = discounted_rewards - value
		policy_loss = - logp * tf.stop_gradient(advantages)

		##### Value loss ######
		value_loss = self.value_reg * tf.square(advantages)

		##### Entropy #####
		entropy = - self.entropy_reg * tf.reduce_sum(policy * tf.log(policy + 1e-10), axis=1, keep_dims=True)

		return tf.reduce_mean(value_loss + policy_loss + entropy)

	def train(self, batch_size, number_of_sequences):
		batch = self.agent.get_memory().get_batches(batch_size, number_of_sequences)
		end_states = (1 - np.asarray(batch.get_dones()).astype(int))
		discounted_rewards = self.discount(batch.get_rewards(), self.agent.get_discount_factor()) * end_states

		starting_states = np.asarray(batch.get_starting_states()).reshape(-1, self.network.get_state_dimension())
		
		grads = self.agent.get_sess().run(self.network.optimizer, feed_dict={
			self.network.y_: discounted_rewards,
			self.network.actions: batch.get_actions(),
			self.network.x: starting_states})
		return

	def columns_initializer(self, std=0.1):
		def _initializer(shape, dtype=None, partition_info=None):
			out = np.random.randn(*shape).astype(np.float32)
			out *= std / np.sqrt(np.square(out).sum(axis=0, keepdims=True))
			return tf.constant(out)
		return _initializer



from __future__ import absolute_import

import tensorflow as tf

from skopos.utils.preprocessing import Preprocessing 

class Network(object):
	"""docstring for NetworkBuilder"""
	def __init__(self, scope="main"):
		super(Network, self).__init__()
		self.layers = []
		self.env = None
		self.trainer = None
		self.learner = None
		self.scope = scope
		self.distributed_training = False

	def initialize_environment_variables(self):
		""" The possible observation spaces could be both Images or Vectors """
		self.act_dimension = self.env.action_space.n
		try:
			self.img_height = self.env.observation_space.shape[0]
			self.img_width = self.env.observation_space.shape[1]
			self.img_input_channels = self.env.observation_space.shape[2]
			
			self.state_shape = [self.img_height, self.img_width, self.img_input_channels]
			self.state_dimension = self.env.observation_space.shape[0]*self.env.observation_space.shape[1]*self.env.observation_space.shape[2]
		except IndexError:
			self.state_dimension = self.env.observation_space.shape[0]
		return

	def inference(self, x, reuse=False):
		for i, layer in enumerate(self.layers):
			x = self.check_input_layer(self.layers, i, x)
			with tf.variable_scope(self.scope + layer.get_scope() + str(i), reuse=reuse):
				x = layer.apply_layer(x)
				try:
					self.set_state_shape([x.shape[1], x.shape[2], x.shape[3]])
				except IndexError:
					pass
		x = Preprocessing.from_tensor_to_vector(x)
		x = self.learner.output(x, reuse)
		return x

	def prediction(self, out):
		return self.learner.prediction(out)

	def error(self, y, y_, actions):
		return self.learner.error_function(y, y_, actions)

	def optimize(self, loss):
		if self.distributed_training == True:
			local_vars = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, self.get_scope())
			gradients = tf.gradients(loss, local_vars)
			var_norms = tf.global_norm(local_vars)
			""" Why 40.0? """
			grads, grad_norms = tf.clip_by_global_norm(gradients, 40.0)
			global_vars = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, "main")
			return self.trainer.apply_gradients(zip(grads, global_vars), name="optimizer")
		else:
			return self.trainer.minimize(loss, name="optimizer")

	def build_graph(self):
		self.initialize_environment_variables()
		self.x, self.y_, self.actions = self.learner.define_placeholders()
		self.out = self.inference(self.x, reuse=False)
		self.loss = self.error(self.out, self.y_, self.actions)
		self.optimizer = self.optimize(self.loss)

	""" User methods to build the initial network. """
	def add(self, layer):
		self.layers.append(layer)

	def set_optimizer(self, trainer):
		self.trainer = trainer.initialize_optimizer()

	""" Module methods to manage the network """
	def get_trainer(self):
		return self.trainer

	def set_trainer(self, trainer):
		self.trainer = trainer

	def get_layers(self):
		return self.layers

	def set_distributed_training(self, distributed_training):
		self.distributed_training = distributed_training

	def get_distributed_training(self):
		return self.distributed_training

	def set_layers(self, layers):
		self.layers = layers

	def set_learner(self, learner):
		self.learner = learner

	def get_learner(self):
		return self.learner

	def get_agent(self):
		return agent

	def set_environment(self, env):
		self.env = env

	def get_environment(self):
		return self.env

	def get_action_number(self):
		return self.act_dimension

	def get_state_dimension(self):
		return self.state_dimension

	def get_img_height(self):
		try:
			return self.img_height
		except AttributeError:
			return

	def get_img_width(self):
		try:
			return self.img_width
		except AttributeError:
			return

	def get_img_input_channels(self):
		try:
			return self.img_input_channels
		except AttributeError:
			return
	def get_state_shape(self):
		return self.state_shape

	def set_state_shape(self, shape):
		self.state_shape = shape

	def get_scope(self):
		return self.scope

	def check_input_layer(self, layers, i, out):
		if layers[i].get_scope() == "conv":
			out = Preprocessing.from_vector_to_image(out, self.get_state_shape())
			return out
		elif layers[i].get_scope() == "max_pool":
			out = Preprocessing.from_vector_to_image(out, self.get_state_shape())
			return out
		else:
			return Preprocessing.from_tensor_to_vector(out)

	









		





from __future__ import absolute_import

import tensorflow as tf
import numpy as np

from skopos.network.network_builder import Network

class NetProcessing(object):
	"""docstring for NetProcessing"""
	def __init__(self):
		super(NetProcessing, self).__init__()

	@staticmethod
	def copy_network(network, scope):
		new_network = Network(scope=scope)
		new_network.set_layers(network.get_layers())
		new_network.set_trainer(network.get_trainer())
		new_network.set_distributed_training(network.get_distributed_training())
		new_network.set_environment(network.get_environment())
		return new_network

	@staticmethod
	def update_graph(from_scope, to_scope):
		from_vars = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, from_scope)
		to_vars = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, to_scope)
		op_holder = []
		for from_var, to_var in zip(from_vars, to_vars):
			op_holder.append(to_var.assign(from_var))
		return op_holder


from __future__ import absolute_import

import tensorflow as tf
import numpy as np

from skopos.network.simple_layers import FullyConnected
from skopos.utils.preprocessing import Preprocessing
from skopos.techniques.learners import ValueIterationLearner

class Dueling(ValueIterationLearner):
	"""docstring for Dueling"""
	def __init__(self, double_network=None):
		super(Dueling, self).__init__()
		self.double_network = double_network

	def define_placeholders(self):
		x = tf.placeholder(np.float32, shape=[None, self.network.get_state_dimension()])
		y_ = tf.placeholder(np.float32, shape=[None])
		actions = tf.placeholder(shape=[None], dtype=tf.int32)
		return x, y_, actions

	def output(self, x, reuse):
		stream = tf.contrib.layers.flatten(x)
		stream_a, stream_v = tf.split(x, 2, 1) 

		advatage_layer = FullyConnected(size=self.network.get_action_number())
		with tf.variable_scope(self.network.get_scope() + 'advatage_layer', reuse=reuse):
			advantage = advatage_layer.apply_layer(stream_a)

		value_layer = FullyConnected(size=1)
		with tf.variable_scope(self.network.get_scope() + 'value_layer', reuse=reuse):
			value = value_layer.apply_layer(stream_v)

		out = value + tf.subtract(advantage, tf.reduce_mean(advantage, axis=1, keep_dims=True))
		return out

	def prediction(self, out):
		return np.argmax(out, 1)

	def error_function(self, y, y_, actions):
		actions_ohe = tf.one_hot(actions, self.network.get_action_number(), dtype=tf.float32)
		q = tf.reduce_sum(y * actions_ohe, reduction_indices=1)
		return tf.reduce_mean(tf.square(y_ - q), name="loss")

	def train(self, batch_size, number_of_sequences):
		batch = self.agent.get_memory().get_batches(batch_size, number_of_sequences)
		final_states = np.asarray(batch.get_final_states()).reshape(-1, self.network.get_state_dimension())
		next_step_rewards = self.agent.get_sess().run(
			self.network.out, 
			feed_dict={self.network.x: final_states})
		actions = np.argmax(next_step_rewards, 1)
		end_states = (1 - np.asarray(batch.get_dones()).astype(int))

		try:
			target_q = self.run_double_network()
		except AttributeError:
			max_expected_reward = np.max(next_step_rewards, 1)
			target_q = batch.get_rewards() + self.agent.get_discount_factor() * max_expected_reward * end_states

		starting_states = np.asarray(batch.get_starting_states()).reshape(-1, self.network.get_state_dimension())
		self.agent.get_sess().run(
			self.network.optimizer, 
			feed_dict={self.network.x: starting_states, self.network.y_: target_q, self.network.actions: actions})
		
		return


from __future__ import absolute_import

import tensorflow as tf

class Layer(object):
	"""docstring for Layers"""
	def __init__(self):
		super(Layer, self).__init__()

	@staticmethod
	def get_variables(name, shape, weight_decay=None, initializer=tf.contrib.layers.xavier_initializer()):
		""" Getting a variable """
		if weight_decay is not None:
			regularizer = tf.contrib.layers.l2_regularizer(scale=weight_decay)
			new_variables = tf.get_variable(name, shape=shape, initializer=initializer, regularizer=regularizer)
		else:
			new_variables = tf.get_variable(name, shape=shape, initializer=initializer)
		return new_variables

	def set_network(self, network):
		self.network = network

		



from __future__ import absolute_import

import tensorflow as tf

from skopos.network.layers import Layer

class ImageLayer(Layer):
	"""docstring for ImageLayer"""
	def __init__(self):
		super(ImageLayer, self).__init__()

class Convolutional(ImageLayer):
	"""docstring for Convolutional"""
	def __init__(self, k_size=3, stride=1, filters=16, 
		padding='SAME', bias=False, weight_stddev=0.1, weight_decay=0.00004):
		super(Convolutional, self).__init__()
		self.k_size = k_size
		self.stride = stride
		self.filters = filters
		self.padding = padding
		self.bias = bias
		self.weight_stddev = weight_stddev
		self.weight_decay = weight_decay

	def apply_layer(self, out):
		
		filters_in = out.get_shape()[-1]
		shape = [self.k_size, self.k_size, filters_in, self.filters]
		initializer = tf.truncated_normal_initializer(stddev=self.weight_stddev)
		weights = Layer.get_variables('conv_weights',
	                            shape=shape,
	                            weight_decay=self.weight_decay,
	                            initializer=initializer)
		if self.bias == True:
			bias = Layer.get_variables('conv_bais',
		   							shape=[filters],
		   							initializer=initializer)
		out = tf.nn.conv2d(out, weights, [1, self.stride, self.stride, 1], self.padding)
		return out

	def get_scope(self):
		return "conv"

class MaxPool(ImageLayer):
	"""docstring for MaxPool"""
	def __init__(self, k_size=3, stride=2, padding='SAME'):
		super(MaxPool, self).__init__()
		self.k_size = k_size
		self.stride = stride
		self.padding = padding

	def apply_layer(self, out):
		out = tf.nn.max_pool(out,
	                          ksize=[1, self.k_size, self.k_size, 1],
	                          strides=[1, self.stride, self.stride, 1],
	                          padding='SAME')
		return out

	def get_scope(self):
		return "max_pool"
		


import tensorflow as tf
import numpy as np
from skopos.network.layers import Layer

class RecurrentLayer(Layer):
	"""docstring for RecurrentLayer"""
	def __init__(self):
		super(RecurrentLayer, self).__init__()

class LSTM(RecurrentLayer):
	"""docstring for LSTM"""
	def __init__(self, size=100):
		super(LSTM, self).__init__()
		self.size = size
		self.initialize_LSTM_cell()

	def initialize_LSTM_cell(self):
		with tf.variable_scope(self.get_scope(), reuse=False):
			"""
				Create a LSTM cell with UNITS_LSTM units. We then define an initial
				state for the LSTM cell. The initial state is divided into 'c' and 'h',
				each with dimension [LSTM_BATCH_SIZE, LSTM_CELL_STATE_SIZE], where
				LSTM_BATCH_SIZE=1. We use 'self.lstm_state_init' only when we want
				te reset the LSTM layer's state (between episodes, for example)
			"""
			self.cell = tf.contrib.rnn.BasicLSTMCell(num_units=self.size, state_is_tuple=True)
			c_init = np.zeros((1, self.cell.state_size.c), np.float32)
			h_init = np.zeros((1, self.cell.state_size.h), np.float32)
			self.state_in = tf.contrib.rnn.LSTMStateTuple(c_init, h_init)
		return

	def apply_layer(self, x):
		"""
			Create new dimension, so that the input for the LSTM is (if using the
			fully connected layer):
			[batch_size=1, time_step=Conv_layer_batch_size, input_dim=UNITS_H1]
				- lstm_input.get_shape() ===> (1, ?, UNITS_H1)
			or, if not using the fully connected layer, it is:
			[batch_size=1, time_step=Conv_layer_batch_size, input_dim=SIZE_CONV_FLAT]
				- lstm_input.get_shape() ===> (1, ?, SIZE_CONV_FLAT)
		"""
		out = tf.expand_dims(x, [0])
		"""
			Retrives the original batch size (used in the Convolutional layer).
			This value (represented by the ? in 'lstm_input.get_shape()') indicates
			how many steps the LSTM layer will unroll. This makes sense, since the
			current batch carries an ordered sequence of events, which is required
			by the LSTM layer.
		"""
		step_size = tf.shape(x)[:1]
		""" 
			The update of the state is done dinamically, the next state is assigned to state_in, in order
			to avoid to feed the variable manually 
		"""
		self.rnn, self.state_in = tf.nn.dynamic_rnn(inputs=out, cell=self.cell, 
			dtype=tf.float32, initial_state=self.state_in, sequence_length=step_size)
		return tf.reshape(self.rnn, [-1, self.size])

	def get_scope(self):
		return "lstm"

class RNN(RecurrentLayer):
	""" Implementing RNN Layer"""
	def __init__(self, arg):
		super(RNN, self).__init__()
		self.arg = arg
		

		



from __future__ import absolute_import

import tensorflow as tf
import numpy as np
import scipy.signal

from skopos.network.simple_layers import FullyConnected
from skopos.utils.preprocessing import Preprocessing
from skopos.utils.net_processing import NetProcessing

class Learner(object):
	""" A learner is a Technique used to predict and learn the best action. """
	def __init__(self):
		super(Learner, self).__init__()
	
	def set_network(self, network):
		self.network = network

	def set_agent(self, agent):
		self.agent = agent

	def discount(self, x, gamma):
		return scipy.signal.lfilter([1], [1, -gamma], x[::-1], axis=0)[::-1] 

class ValueIterationLearner(Learner):
	"""docstring for ClassName"""
	def __init__(self):
		super(ValueIterationLearner, self).__init__()

	def initialize_double_network():
		target_network = NetProcessing.copy_network(self.network, "target")
		target_network.build_graph()
		target_network.set_learner(self)
		return target_network

	def run_double_network(self):
		if self.double_network is not None and self.double_network.get_network() is not None:
			self.double_network.set_network(self.initialize_double_network()) 
		""" Updates the variables of the target network using the variables of the main one """
		if self.double_network.get_total_updates() % self.double_network.get_tau() == 0:
			self.agent.get_sess().run(NetProcessing.update_graph(self.network.get_scope(), to_scope=self.double_network.get_network().get_scope()))
		final_states = np.asarray(batch.get_final_states()).reshape(-1, self.network.get_state_dimension())
		double_q = self.agent.get_sess().run(
			self.target_network.out, 
			feed_dict={self.target_network.x: final_states})
		max_double_q = double_q[range(batch_size), actions]
		target_q = batch.get_rewards() + self.agent.get_discount_factor() * max_double_q * end_states
		return target_q

class PolicyGradientLearner(Learner):
	"""docstring for ClassName"""
	def __init__(self):
		super(PolicyGradientLearner, self).__init__()

class ValuePolicyLearner(Learner):
	"""docstring for ClassName"""
	def __init__(self):
		super(ValuePolicyLearner, self).__init__()


		
		



from __future__ import absolute_import

import numpy as np
import random

class ExplorationStrategy(object):
	"""docstring for ExplorationStrategy"""
	def __init__(self):
		super(ExplorationStrategy, self).__init__()

	def get_environment(self):
		return self.policy.get_environment()

	def evaluate(self, state):
		raise NotImplementedError('Subclasses must override the method.')

	def set_policy(self, policy):
		self.policy = policy

class Greedy(ExplorationStrategy):
	"""docstring for Greedy"""
	def __init__(self):
		super(Greedy, self).__init__()

	def evaluate(self, state):
		action, expected_rewards = self.policy.evaluate(state)
		return action, expected_rewards 

class EGreedy(ExplorationStrategy):
	
	def __init__(self, epsilon=0.1):
		super(EGreedy, self).__init__()
		self.epsilon = epsilon

	def evaluate(self, state):
		action, expected_rewards = self.policy.evaluate(state)
		if random.random < self.epsilon:
			action = random.randint(0, self.env.action_space.n - 1)
			return action, expected_rewards
		else:
			return action, expected_rewards

class DecrementalEGreedy(ExplorationStrategy):
	
	def __init__(self, epsilon=0.4, discountFactor=0.99, minimumEpsilon=0.1):
		self.epsilon = epsilon
		self.discountFactor = discountFactor
		self.minimumEpsilon = minimumEpsilon

	def evaluate(self, state):
		action, expected_rewards = self.policy.evaluate(state)
		condition = random.random < self.epsilon
		self.epsilon = max(self.minimumEpsilon, self.epsilon * self.discountFactor)

		if condition:
			action = random.randint(0, self.env.action_space.n - 1)
			return action, expected_rewards
		else:
			return action, expected_rewards

class Boltzman(ExplorationStrategy):
	
	def __init__(self, epsilon=0.4, discountFactor=0.99, minimumEpsilon=0.1):
		self.epsilon = epsilon
		self.discountFactor = discountFactor
		self.minimumEpsilon = minimumEpsilon

	def evaluate(self, state):
		action, expected_rewards_ = self.policy.evaluate(state)
		expected_rewards = np.squeeze(expected_rewards_) 
		expected_rewards = expected_rewards + np.absolute(min(expected_rewards))
		boltzman_values = - expected_rewards / self.epsilon
		boltzman_values = [np.exp(value) for value in boltzman_values]
		boltzman_values_sum = sum(boltzman_values)
		probabilities = []

		for element in boltzman_values:
			prob = element / boltzman_values_sum
			probabilities.append(prob)


		probabilities = np.squeeze(probabilities)
		chosen_value = np.random.choice(expected_rewards, p=probabilities)
		self.epsilon = max(self.minimumEpsilon, self.epsilon * self.discountFactor)	

		return np.argmax(expected_rewards == chosen_value), expected_rewards_



		




from __future__ import absolute_import

import tensorflow as tf
import numpy as np

from skopos.memories.memory import ExperienceReplay, Episode
from skopos.utils.visualize import Plot
from skopos.agent.exploration_strategy import EGreedy
from skopos.agent.policies import NetworkBasedPolicy
from skopos.agent.environment_preprocessing import Environment
from skopos.agent.agent import Agent
from skopos.techniques.dqn import DQN

class SimpleAgent(Agent):

	def __init__(self, env, network, processors=4, number_of_episodes=1000, 
		max_episode_duration=100, discount_factor=0.99, pretrain_steps=100,
		batch_size=32, update_frequency=1, sequences=1, 
		training_info=False, show_results=False, tensorboard_visualization=False,
		training_parameters=False, 
		save_model=False, model_folder="./model/", model_path="./model/model", 
		save_model_frequency=50, restore_model=False, log_dir="./tensorboard",
		learner=DQN(), exploration_strategy=EGreedy(), 
		policy=NetworkBasedPolicy(), memory=ExperienceReplay()):
		super(SimpleAgent, self).__init__()
		""" Building the computation graph for running the algo """
		self.memory = memory
		self.network = network
		self.learner = learner
		""" Sharing references """
		self.env = env
		self.network.set_environment(self.env)
		self.network.set_learner(self.learner)
		self.learner.set_network(self.network)
		self.env_preprocessing = Environment(self.env)
		self.processors = processors
		self.initialize_tensorflow_computation(self.processors)
		""" Running parameters """
		self.number_of_episodes = number_of_episodes
		self.max_episode_duration = max_episode_duration
		self.discount_factor = discount_factor
		self.pretrain_steps = pretrain_steps
		self.update_frequency = update_frequency
		self.batch_size = batch_size
		self.sequences = sequences
		""" Policy and Exploration Strategy parameters """
		self.exploration_strategy = exploration_strategy
		self.learner.set_agent(self)
		self.policy = policy
		self.initialize_policy()
		""" Define general parameters """
		self.training_info = training_info
		self.training_parameters = training_parameters
		self.show_results = show_results
		self.tensorboard_visualization = tensorboard_visualization
		self.save_model = save_model
		self.restore_model = restore_model
		self.model_folder = model_folder
		self.model_path = model_path
		self.save_model_frequency = save_model_frequency
		self.log_dir = log_dir
		if self.tensorboard_visualization == True:
			self.summary_writer = tf.summary.FileWriter(self.log_dir + "/train")

	def initialize_policy(self):
		self.policy.set_env(self.env)
		self.policy.set_network(self.network)
		self.policy.set_session(self.sess)
		self.exploration_strategy.set_policy(self.policy)
		return

	def initialize_tensorflow_computation(self, processors):
		self.network.build_graph()
		self.config = tf.ConfigProto(intra_op_parallelism_threads=processors, inter_op_parallelism_threads=processors)
		self.sess = tf.Session(config=self.config)
		self.sess.run(tf.global_variables_initializer())
		self.saver = tf.train.Saver()
		return

	def run(self):

		if self.training_parameters == True:
			Plot.show_training_parameters(self)

		self.restore_model_from_ckpt()
		self.reward_list = []
		self.episodes_length = []
		self.avg_max_exp_rewards = []
		self.total_steps = 0

		for i in range(0, self.number_of_episodes):
			""" Processing the state in order to feed it into the tensorflow network """
			s = self.env.reset()
			s = self.env_preprocessing.process_state(s)

			episode_reward = 0
			episode_avg_exp_rewards = 0
			done = False
			episode = Episode()

			for j in range(0, self.max_episode_duration):
				self.total_steps = self.total_steps + 1

				if self.total_steps < self.pretrain_steps:
					""" Pretraining to collect experience """
					self.pretrain_steps = self.pretrain_steps - 1
					action = np.random.randint(0, self.env.action_space.n)
					expected_rewards = np.ones(self.env.action_space.n)
				else:
					""" Predicting function """
					action, expected_rewards = self.exploration_strategy.evaluate(s)
				
				landing_state, reward, done, info = self.env.step(action)
				landing_state = self.env_preprocessing.process_state(landing_state)

				""" Record statistics """
				episode.record_step(s, action, reward, landing_state, done, expected_rewards)
				episode_reward += reward
				episode_avg_exp_rewards += np.max(expected_rewards)
				s = landing_state

				if done == True:
					if self.training_info == True:
						Plot.visualize_training(j, episode_reward)
					if self.tensorboard_visualization == True:
						self.summary = tf.Summary()
						self.summary.value.add(tag='Reward', simple_value=float(np.mean(self.reward_list[-5:])))
						self.summary_writer.add_summary(self.summary, i)
						self.summary_writer.flush()
					break

			if i % self.save_model_frequency == 0 and self.save_model == True:
				self.save_model_ckpt()

			self.reward_list.append(episode_reward)
			self.episodes_length.append(j)
			self.avg_max_exp_rewards.append(episode_avg_exp_rewards/j)
			self.memory.record_episode(episode)

			if i % self.update_frequency == 0 and self.total_steps > self.pretrain_steps:
				self.learner.train(self.batch_size, self.sequences)

		if self.show_results == True:
			Plot.print_statistics(self.memory)
			Plot.visualize_results(self.reward_list, self.episodes_length, self.avg_max_exp_rewards)
		return self.reward_list



from __future__ import absolute_import

import tensorflow as tf
import numpy as np
import threading
import multiprocessing
from time import sleep
import copy
import Queue

""" Default technique """
from skopos.techniques.dqn import DQN

""" Default memory """
from skopos.memories.memory import ExperienceReplay, Episode

""" Utils for visualization and preprocessing """
from skopos.utils.visualize import Plot
from skopos.utils.net_processing import NetProcessing

""" Default exploration strategy """
from skopos.agent.exploration_strategy import EGreedy

""" Default policy """
from skopos.agent.policies import NetworkBasedPolicy
from skopos.agent.environment_preprocessing import Environment
from skopos.agent.agent import Agent
from skopos.network.network_builder import Network


class Worker(Agent):
	"""docstring for Worker"""
	def __init__(self, agent, name, env):
		super(Worker, self).__init__()
		self.name = "worker_" + str(name)
		self.env = env
		self.copy_agent(agent)
		self.saver = tf.train.Saver()
		if self.tensorboard_visualization == True:
			self.summary_writer = tf.summary.FileWriter(self.log_dir + "/train_" + str(name))
		print("%s, created!" % self.name)

	def copy_agent(self, agent):
		self.memory = type(agent.get_memory())() 
		self.env_preprocessing = Environment(self.env)
		self.sess = agent.get_sess()
		self.learner = type(agent.get_learner())()
		self.local_network = NetProcessing.copy_network(agent.get_network(), scope=self.name)
		self.local_network.set_distributed_training(True)
		self.local_network.set_learner(self.learner)
		self.learner.set_network(self.local_network)
		""" General parameters """
		self.training_info = agent.get_training_info()
		self.tensorboard_visualization = agent.get_tensorboard_visualization()
		self.save_model = agent.get_save_model()
		self.restore_model = agent.get_restore_model()
		self.model_folder = agent.get_model_folder()
		self.model_path = agent.get_model_path()
		self.save_model_frequency = agent.get_save_model_frequency()
		""" Running parameters """
		self.number_of_episodes = agent.get_number_of_episodes()
		self.max_episode_duration = agent.get_max_episode_duration()
		self.discount_factor = agent.get_discount_factor()
		self.pretrain_steps = agent.get_pretrain_steps()
		self.update_frequency = agent.get_update_frequency()
		self.batch_size = agent.get_batch_size()
		self.sequences = agent.get_sequences()
		self.processors = agent.get_processors()
		""" Policy and Exploration Strategy parameters """
		self.exploration_strategy = type(agent.get_exploration_strategy())()
		self.learner.set_agent(self)
		self.policy = type(agent.get_policy())()
		self.initialize_policy()
		self.local_network.build_graph()
		self.log_dir = agent.get_log_dir()
		return 

	def initialize_policy(self):
		self.policy.set_env(self.env)
		self.policy.set_network(self.local_network)
		self.policy.set_session(self.sess)
		self.exploration_strategy.set_policy(self.policy)
		return

	def work(self, T_queue):

		if self.name == "worker_0":
			self.restore_model_from_ckpt()
		
		T = T_queue.get()
		T_queue.put(T+1)
		self.reward_list = []
		self.episodes_length = []
		self.avg_max_exp_rewards = []
		
		print("Starting worker " + str(self.name))
		while T < self.number_of_episodes * self.processors:
			try:
				""" In case still doesnt exit yet """
				self.sess.run(NetProcessing.update_graph(from_scope="main", to_scope=self.name))
			except ValueError:
				pass
			""" Processing the state in order to feed it into the tensorflow network """
			s = self.env.reset()
			s = self.env_preprocessing.process_state(s)

			episode_reward = 0
			episode_avg_exp_rewards = 0
			done = False
			episode = Episode()
			episode_steps = 0
			
			for j in range(0, self.max_episode_duration):

				episode_steps = episode_steps + 1

				if T < self.pretrain_steps:
					""" Pretraining to collect experience """
					self.pretrain_steps = self.pretrain_steps - 1
					action = np.random.randint(0, self.env.action_space.n)
					expected_rewards = np.ones(self.env.action_space.n)
				else:
					""" Predicting function """
					action, expected_rewards = self.exploration_strategy.evaluate(s)
				
				landing_state, reward, done, info = self.env.step(action)
				landing_state = self.env_preprocessing.process_state(landing_state)
				""" Updating global counter """
				T = T_queue.get()
				T_queue.put(T+1)

				""" Record statistics """
				episode.record_step(s, action, reward, landing_state, done, expected_rewards)
				episode_reward += reward
				episode_avg_exp_rewards += np.max(expected_rewards)
				s = landing_state

				if done == True:
					if self.training_info == True:
						Plot.visualize_training(j, episode_reward)
					if self.tensorboard_visualization == True:
						self.summary = tf.Summary()
						self.summary.value.add(tag='Reward', simple_value=float(np.mean(self.reward_list[-5:])))
						self.summary_writer.add_summary(self.summary, T)
						self.summary_writer.flush()
					break

			if self.name == "worker_0" and T % self.save_model_frequency == 0 and self.save_model == True:
				self.save_model_ckpt()

			self.reward_list.append(episode_reward)
			self.episodes_length.append(j)
			self.avg_max_exp_rewards.append(episode_avg_exp_rewards/j)
			self.memory.record_episode(episode)
			if T % self.update_frequency == 0 and T > self.pretrain_steps:
				self.learner.train(self.batch_size, self.sequences)

		return

class AsynchronousAgent(Agent):
	""" The init function takes in input the Tensorflow parameters """
	def __init__(self, env, network, processors=4, number_of_episodes=1000, 
		max_episode_duration=100, discount_factor=0.99, pretrain_steps=100,
		batch_size=32, update_frequency=1, sequences=1, 
		training_info=False, training_parameters=False,
		show_results=False, tensorboard_visualization=False, save_model=False,
		model_folder="./model/", model_path="./model/model", 
		save_model_frequency=50, restore_model=False, log_dir="./tensorboard",
		learner=DQN(), exploration_strategy=EGreedy(), 
		policy=NetworkBasedPolicy(), memory=ExperienceReplay()):
		super(AsynchronousAgent, self).__init__()
		""" Building the computation graph for running the algo """
		self.memory = memory
		self.network = network
		self.learner = learner
		""" Sharing references """
		self.env = env
		self.network.set_environment(self.env)
		self.network.set_learner(self.learner)
		self.network.set_distributed_training(True)
		self.learner.set_network(self.network)
		self.env_preprocessing = Environment(self.env)
		""" Running parameters """
		self.number_of_episodes = number_of_episodes
		self.max_episode_duration = max_episode_duration
		self.discount_factor = discount_factor
		self.pretrain_steps = pretrain_steps
		self.update_frequency = update_frequency
		self.batch_size = batch_size
		self.sequences = sequences
		""" Define general parameters """
		self.training_info = training_info
		self.training_parameters = training_parameters
		self.show_results = show_results
		self.tensorboard_visualization = tensorboard_visualization
		self.save_model = save_model
		self.restore_model = restore_model
		self.processors = processors
		self.model_folder = model_folder
		self.model_path = model_path
		self.save_model_frequency = save_model_frequency
		self.log_dir = log_dir
		""" Policy and Exploration Strategy parameters """
		self.exploration_strategy = exploration_strategy
		self.learner.set_agent(self)
		self.policy = policy
		self.sess = tf.Session()
		self.initialize_policy()
		

	def initialize_policy(self):
		self.policy.set_env(self.env)
		self.policy.set_network(self.network)
		self.policy.set_session(self.sess)
		self.exploration_strategy.set_policy(self.policy)
		return

	def run(self):
		if self.training_parameters == True:
			Plot.show_training_parameters(self)
		
		with tf.device("/cpu:0"):
			print("Global Network Initialization...")
			self.network.build_graph()
			self.sess.run(tf.global_variables_initializer())
			print("Global Network Initialized...")
			print("Creating the global counter...")
			T_queue = Queue.Queue()
			T_queue.put(0)
			self.workers = []
			print("Instantiating all the workers...")
			for i in range(0, self.processors):
				self.workers.append(Worker(self, i, env=copy.deepcopy(self.env)))
		
		self.coord = tf.train.Coordinator()
		self.sess.run(tf.global_variables_initializer())
		
		processes = []
		for worker in self.workers:
			worker_work = lambda: worker.work(T_queue=T_queue)
			t = threading.Thread(target=(worker_work))
			t.start()
			processes.append(t)
		self.coord.join(processes)

		if self.show_results == True:
			Plot.visualize_results(self.workers[0].reward_list, self.workers[0].episodes_length, self.workers[0].avg_max_exp_rewards)
		return


















from __future__ import absolute_import
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import numpy as np

class Plot(object):
	"""docstring for Plot"""
	def __init__(self, arg):
		super(Plot, self).__init__()
		self.arg = arg

	@staticmethod
	def visualize_results(reward_list, episodes_length, avg_max_exp_rewards):
		gs = gridspec.GridSpec(1, 3
			)
		plt.figure(1, figsize=(12, 6))
		ax = plt.subplot(gs[0, 0])
		ax.set_title("Rewards per episode")
		ax.plot(reward_list)

		ax = plt.subplot(gs[0, 1])
		ax.set_title("Number of actions (length) per episode")
		ax.plot(episodes_length)

		ax = plt.subplot(gs[0, 2])
		ax.set_title("Max expected reward per episode")
		ax.plot(avg_max_exp_rewards)

		plt.show()
		return

	@staticmethod
	def visualize_training(step, episode_reward):
		print("Game finished after %s iterations. With an avg reward of %s." %(step, episode_reward/step))
		return

	@staticmethod
	def print_statistics(experience):
		a, b, c, d = experience.get_statistics()
		print("{} played games for a total of {} action taken. In average I score {} per action for a total average reward per episode of {}." .format(a,b,c,d))

	@staticmethod
	def show_training_parameters(agent):
		print("Used processors during the training: ", agent.get_processors())
		print("Number of episodes: ", agent.get_number_of_episodes())
		print("Max episode duration: ", agent.get_max_episode_duration())
		print("Discount factor (lambda for value iteration learners and dropping rate for computing discounted rewards): ", agent.get_discount_factor())
		print("Number of Pre Train steps done randomly to collect a certain experience: ", agent.get_pretrain_steps())
		print("Frequency update (how often compute an optimization step): ", agent.get_update_frequency())
		
		print("Batch size (number of episodes): ", agent.get_batch_size())
		print("Number of consecutive experiences in the episode to consider (RNN, Eligibity trace): ", agent.get_sequences())
		
		print("Save the model? ", agent.get_save_model())
		print("Folder path where to save the model ckpt: ", agent.get_model_folder())
		print("Restore the model? ", agent.get_restore_model())
		print("Model path where to restore the model: ", agent.get_model_path())
		print("How often save the model: ", agent.get_save_model_frequency())
		print("Tensorboard directory: ", agent.get_log_dir())
		
		print("Learner used for training: ", agent.get_learner())
		print("Memory used for storing experiences: ", agent.get_memory())
		print("Exploration strategy used for training: ", agent.get_exploration_strategy())
		print("Policy used for training: ", agent.get_policy())



from __future__ import absolute_import

import tensorflow as tf
import numpy as np

class Preprocessing(object):
	"""docstring for Preprocessing"""
	def __init__(self, arg):
		super(Preprocessing, self).__init__()
		self.arg = arg

	@staticmethod
	def from_vector_to_image(x, shape):
		return tf.reshape(x, shape=[-1, shape[0], shape[1], shape[2]])

	@staticmethod
	def from_tensor_to_vector(x):
		return tf.contrib.layers.flatten(x)


from __future__ import absolute_import
import tensorflow as tf

class Optimizer(object):
	"""docstring for Optimizer"""
	def __init__(self):
		super(Optimizer, self).__init__()

class Adam(Optimizer):
	"""docstring for Adam"""
	def __init__(self, learning_rate):
		super(Adam, self).__init__()
		self.learning_rate = learning_rate

	def initialize_optimizer(self):
		return tf.train.AdamOptimizer(learning_rate = self.learning_rate)

class Gradient(Optimizer):
	"""docstring for Adam"""
	def __init__(self, learning_rate):
		super(Gradient, self).__init__()
		self.learning_rate = learning_rate

	def initialize_optimizer(self):
		return tf.train.GradientDescentOptimizer(learning_rate = self.learning_rate)

class RMSProp(Optimizer):
	"""docstring for Adam"""
	def __init__(self, learning_rate):
		super(RMSProp, self).__init__()
		self.learning_rate = learning_rate

	def initialize_optimizer(self):
		return tf.train.RMSPropOptimizer(learning_rate = self.learning_rate)


		
	


from __future__ import absolute_import

import tensorflow as tf
import numpy as np

from skopos.network.simple_layers import FullyConnected
from skopos.utils.preprocessing import Preprocessing
from skopos.techniques.learners import ValueIterationLearner

class NStepQLearning(ValueIterationLearner):

	def __init__(self):
		super(NStepQLearning, self).__init__()

	def define_placeholders(self):
		x = tf.placeholder(tf.float32, shape=[None, self.network.get_state_dimension()])
		y_ = tf.placeholder(tf.float32, shape=[None])
		actions = tf.placeholder(tf.int32, shape=[None])
		return x, y_, actions

	def output(self, x, reuse):
		layer = FullyConnected(size=self.network.get_action_number())
		with tf.variable_scope(self.network.get_scope() + "output_layer", reuse=reuse):
			out = layer.apply_layer(x)
		return out

	def prediction(self, out):
		return np.argmax(out, 1)

	def error_function(self, y, y_, actions):
		actions_ohe = tf.one_hot(actions, self.network.get_action_number(), dtype=tf.float32)
		q = tf.reduce_sum(y * actions_ohe, reduction_indices=1)
		return tf.reduce_mean(tf.square(y_ - q), name="loss")

	def train(self, batch_size, number_of_sequences):
		batch = self.agent.get_memory().get_batches(batch_size, number_of_sequences)
		end_states = (1 - np.asarray(batch.get_dones()).astype(int))
		target_q = self.discount(batch.get_rewards(), self.agent.get_discount_factor()) * end_states
		
		starting_states = np.asarray(batch.get_starting_states()).reshape(-1, self.network.get_state_dimension())
		
		self.agent.get_sess().run(
			self.network.optimizer, 
			feed_dict={
				self.network.x: starting_states, 
				self.network.y_: target_q, 
				self.network.actions: batch.get_actions()})
		
		return


# coding=utf-8

from __future__ import absolute_import
import random
import numpy as np


class Memory(object):
	"""Superclass including the basic and static methods for the memory utilties"""
	def __init__(self):
		super(Memory, self).__init__()

	""" Methods to extract the index of the list to be replaced"""
	def random_replace(number_of_episodes):
		""" Just returns a random number within list size """
		return random.randint(0, number_of_episodes)

	def fifo_replace(number_of_episodes):
		""" Return the index of the oldest element in the list """
		return 0

	def lifo_replace(number_of_episodes):
		""" Return the index of the latest element added in the list """
		return number_of_episodes - 1

	""" Methods to compute the priority level of an episode """
	def constant_priority(episode):
		""" Assigns constant priority to all the episodes (no prioritization) """
		return 1.0

	def tot_reward_difference_priority(episode):
		""" Assigns as priority the sum of the difference between expected and obtained 
			rewards at each step of the episode """
		diff = 0
		for index in range(0, episode.length()):
			choice = episode.action_sequences[index]
			diff += np.absolute(episode.reward_sequences[index] - np.squeeze(episode.expected_rewards[index])[choice])

		return diff / episode.length()

	def max_reward_difference_priority(episode):
		""" Assigns as priority the max difference between expected and obtained 
			rewards of the episode """
		diff = 0
		for index in range(0, episode.length()):
			choice = episode.action_sequences[index]
			diff = max(np.absolute(episode.reward_sequences[index] - episode.expected_rewards[index][choice]), diff)

		return diff

	def undecided_cases_priority(episode):
		""" Measures the indecision of the outuput of the policy by checking
			wether the second optimal choice is close in value to the first, the third 
			to the second and so on and so forth """
		var = 0
		for index in range(0, episode.length):
			ex_rew = list(episode.expected_rewards[index])
			length = 0
			ex_rew = ex_rew.sort()
			for i in range(1, length):
				if ex_rew[i] != 0:
					var += (ex_rew[i-1] / ex_rew[i]) * i
				length += i

			if length != 0:
				var = var / length

		return var

	# Dictionary mapping names of replacement strategies to the respective function
	replacement_strategies = {'random' : random_replace, 'fifo' : fifo_replace, 'lifo' : lifo_replace}
	# Dictionary mapping names of prioritization strategies to the respective function
	prioritization_strategies = {'constant' : constant_priority, 'tot_td' : tot_reward_difference_priority, 'max_td' : max_reward_difference_priority, 'indecision' : undecided_cases_priority}


class ExperienceReplay(Memory):
	""" Class to store the experience """
	def __init__(self, capacity=-1, replacement_strategy='random'):
		super(ExperienceReplay, self).__init__()
		self.episodes = []
		self.episode_number = 0
		self.total_steps = 0
		self.average_total_reward = 0
		self.capacity = capacity
		self.replacement_strategy = Memory.replacement_strategies[replacement_strategy]
		self.priorities = []
		self.prioritization_strategy = Memory.prioritization_strategies['constant']

	def record_episode(self, episode):
		if self.capacity == self.total_steps:
			to_be_replaced = replacement_strategy(self.episode_number)
			self.episodes.pop(to_be_replaced)
			self.priorities.pop(to_be_replaced)
		self.episodes.append(episode)
		self.episode_number += 1
		self.total_steps += episode.length()
		rew_sum = sum(episode.reward_sequences)
		self.average_total_reward += (rew_sum - self.average_total_reward) / self.episode_number
		self.priorities.append(self.prioritization_strategy(episode))
		if(self.episode_number==1):
			self.record_episode(episode)

	""" I need a method get_last_episode """
	def get_episode(self, index):
		if index > self.episode_number:
			print("Index too high. Totally available {} episodes.".format(episode_number))
			return
		return self.episodes[index]

	def get_random_episodes(self, number, prioritized=True):
		if number > self.episode_number:
			print("There are not {} episodes. Only {} available.".format(number, episode_number))
			return

		if prioritized :
			episodes = np.random.choice(self.episodes, size=number, p=(self.priorities / sum(self.priorities)))
		else:
			episodes = np.random.choice(self.episodes, size=number)

		return episodes

	def get_sequences(self, number, size, fixed_length=True, prioritized=True):
		sequences = []

		if prioritized :
			priorities = np.asarray(self.priorities) / sum(self.priorities)
		else:
			priorities = None

		priorities = np.squeeze(priorities)
		while(len(sequences) < number):
			episode = np.random.choice(self.episodes, p=priorities)
			seq = episode.get_sequences(size, fixed_length)
			if seq is not None:
				sequences.append(seq)

		return sequences

	def get_batch(self, size):
		batch = StepBatch()
		for i in range(0, size):
			episode = random.randint(0, self.episode_number)
			""" HERE AN ERROR, in case of the extracted random number is 0, it gives index out of range 

			MOREOVER IF THE GAME IS LONG (BEFORE LOOSING OR WIN TAKES A LOT OF TIME, WE DON'T HAVE ANY 
			AT THE MOMENT OF THE FIRST UPDATE 

			"""
			step = random.randint(0, self.episodes[episode-1].length())
			batch.record_step(self.episodes[episode-1].get_step(step-1))
		return batch

	def get_batches(self, number, size, sequential=True, prioritized=True):
		batches = SequenceBatch()

		if prioritized :
			priorities = np.asarray(np.squeeze(self.priorities)) / sum(np.squeeze(self.priorities))
		else:
			priorities = None

		episodes = []

		priorities = np.squeeze(priorities)
		while(batches.length() < number):
			if episodes == []:
				episodes = list(np.random.choice(self.episodes, number, p=priorities))

			episode = episodes.pop(0)
			

			if sequential:
				sequence = episode.get_sequence(size)
			else:
				sequence = episode.get_batch(size)

			if sequence is not None:
				batch = StepBatch(sequential=sequential)
				for step in sequence:
					batch.record_step(step)

				batches.record_sequence(batch)

		return batches

	def get_statistics(self):
		steps_per_episode = float(self.total_steps / self.episode_number)
		reward_per_step = self.average_total_reward / steps_per_episode
		return self.episode_number, self.total_steps, reward_per_step, self.average_total_reward



class Episode:

	def __init__(self):
		self.action_sequences = []
		self.reward_sequences = []
		self.landing_states = []
		self.starting_states = []
		self.dones = []
		self.expected_rewards = []
		self.episode_length = 0 

	def get_final_states(self):
		return np.squeeze(self.landing_states)

	def get_rewards(self):
		return np.squeeze(self.reward_sequences)

	def get_starting_states(self):
		return np.squeeze(self.starting_states)

	def get_actions(self):
		return np.squeeze(self.action_sequences)

	def length(self):
		return np.squeeze(self.episode_length)

	def get_batch(self, size, fixed_length=True):
		size = min(size, episode_length)
		indices = random.sample(range(0, episode_length), size)
		batch = [self.get_step(index) for index in indices]

		if len(batch) != size and fixed_length:
			return None

		return batch

	def get_dones(self):
		return np.squeeze(self.dones)

	def get_sequence(self, size, fixed_length=True):
		if self.episode_length >= size:
			starting_index = random.randint(0, self.episode_length-size)
		else:
			if fixed_length:
				return None
			starting_index = random.randint(0, self.episode_length)

		ending_index = min(starting_index+size, self.episode_length)
		sample = [self.get_step(index) for index in range(starting_index, ending_index)]

		return sample

	def get_step(self, index):
		return self.starting_states[index], self.action_sequences[index], self.reward_sequences[index], self.landing_states[index], self.dones[index]

	def record_step(self, starting_state, action, reward, final_state, done, expected_rewards):
		self.action_sequences.append(action)
		self.reward_sequences.append(reward)
		self.landing_states.append(final_state)
		self.starting_states.append(starting_state)
		self.dones.append(done)
		self.expected_rewards.append(expected_rewards)
		self.episode_length += 1
		return


class StepBatch:

	def __init__(self, sequential=False):
		self.steps = []
		self.tot_steps = 0
		self.sequential = sequential

	def is_sequential(self):
		return self.sequential

	def record_step(self, step):
		self.steps.append(step)
		self.tot_steps += 1

	def get_rewards(self):
		return np.squeeze([step[2] for step in self.steps])

	def get_starting_states(self):
		return np.squeeze([step[0] for step in self.steps])

	def get_final_states(self):
		return np.squeeze([step[3] for step in self.steps])

	def get_actions(self):
		return np.squeeze([step[1] for step in self.steps])

	def get_dones(self):
		return np.squeeze([step[4] for step in self.steps])

	def toList(self):
		return self.steps


class SequenceBatch:

	def __init__(self):
		self.sequences = list()
		self.tot_sequences = 0

	def length(self):
		return self.tot_sequences

	def record_sequence(self, step_batch):
		self.sequences.append(step_batch)
		self.tot_sequences += 1

	def get_rewards(self):
		rewards = [0]
		for sequence in self.sequences:
			try:
				_list = sequence.get_rewards()
				_to_add = list(_list)

			except Exception as e:
				_to_add = []
				_to_add.append(_list)
				
			rewards.append(np.asarray(_to_add))
		
		rewards.pop(0)
		rewards = np.asarray(rewards).flatten()
		return rewards

	def get_starting_states(self):
		starting_states = [0]
		for sequence in self.sequences:
			starting_states.extend(sequence.get_starting_states())		
		starting_states.pop(0)
		return starting_states

	def get_final_states(self):
		final_states = [0]
		for sequence in self.sequences:
			final_states.extend(sequence.get_starting_states())		
		final_states.pop(0)
		return final_states

	def get_actions(self):
		actions = [0]
		for sequence in self.sequences:
			try:
				_list = sequence.get_actions()
				_to_add = list(_list)

			except Exception as e:
				_to_add = []
				_to_add.append(_list)
				
			actions.append(np.asarray(_to_add))
		
		actions.pop(0)
		actions = np.asarray(actions).flatten()
		return actions

	def get_dones(self):
		dones = [0]
		for sequence in self.sequences:
			try:
				_list = sequence.get_dones()
				_to_add = list(_list)

			except Exception as e:
				_to_add = []
				_to_add.append(_list)

			dones.append(np.asarray(_to_add))
		
		dones.pop(0)
		dones = np.asarray(dones).flatten()
		return dones

	def toList(self):
		return self.sequences


class PrioritizedExperienceReplay(ExperienceReplay):
	"""docstring for Memory"""
	def __init__(self, capacity=-1, replacement_strategy='random', prioritization='tot_td'):
		super(ExperienceReplay, self).__init__()
		self.episodes = []
		self.episode_number = 0
		self.total_steps = 0
		self.average_total_reward = 0
		self.capacity = capacity
		self.replacement_strategy = Memory.replacement_strategies[replacement_strategy]
		self.priorities = []
		self.prioritization_strategy = Memory.prioritization_strategies[prioritization]

	def record_episode(self, episode):
		if self.capacity == self.total_steps:
			to_be_replaced = replacement_strategy(self.episode_number)
			self.episodes.pop(to_be_replaced)
			self.priorities.pop(to_be_replaced)
		self.episodes.append(episode)
		self.episode_number += 1
		self.total_steps += episode.length()
		rew_sum = sum(episode.reward_sequences)
		self.average_total_reward += (rew_sum - self.average_total_reward) / self.episode_number
		self.priorities.append(self.prioritization_strategy(episode))
		if(self.episode_number==1):
			self.record_episode(episode)

	""" I need a method get_last_episode """
	def get_episode(self, index):
		if index > self.episode_number:
			print("Index too high. Totally available {} episodes.".format(episode_number))
			return
		return self.episodes[index]

	def get_random_episodes(self, number, prioritized=True):
		if number > self.episode_number:
			print("There are not {} episodes. Only {} available.".format(number, episode_number))
			return

		if prioritized :
			episodes = np.random.choice(self.episodes, size=number, p=(self.priorities / sum(self.priorities)))
		else:
			episodes = np.random.choice(self.episodes, size=number)

		return episodes

	def get_sequences(self, number, size, fixed_length=True, prioritized=True):
		sequences = []

		if prioritized :
			priorities = np.asarray(self.priorities) / sum(self.priorities)
		else:
			priorities = None

		priorities = np.squeeze(priorities)
		while(len(sequences) < number):
			episode = np.random.choice(self.episodes, p=priorities)
			seq = episode.get_sequences(size, fixed_length)
			if seq is not None:
				sequences.append(seq)

		return sequences

	def get_batch(self, size):
		batch = StepBatch()
		for i in range(0, size):
			episode = random.randint(0, self.episode_number)
			""" HERE AN ERROR, in case of the extracted random number is 0, it gives index out of range 

			MOREOVER IF THE GAME IS LONG (BEFORE LOOSING OR WIN TAKES A LOT OF TIME, WE DON'T HAVE ANY 
			AT THE MOMENT OF THE FIRST UPDATE 

			"""
			step = random.randint(0, self.episodes[episode-1].length())
			batch.record_step(self.episodes[episode-1].get_step(step-1))
		return batch

	def get_batches(self, number, size, sequential=True, prioritized=True):
		batches = SequenceBatch()

		if prioritized :
			priorities = np.asarray(np.squeeze(self.priorities)) / sum(np.squeeze(self.priorities))
		else:
			priorities = None

		episodes = []

		priorities = np.squeeze(priorities)
		while(batches.length() < number):
			if episodes == []:
				episodes = list(np.random.choice(self.episodes, number, p=priorities))

			episode = episodes.pop(0)
			

			if sequential:
				sequence = episode.get_sequence(size)
			else:
				sequence = episode.get_batch(size)

			if sequence is not None:
				batch = StepBatch(sequential=sequential)
				for step in sequence:
					batch.record_step(step)

				batches.record_sequence(batch)

		return batches

	def get_statistics(self):
		steps_per_episode = float(self.total_steps / self.episode_number)
		reward_per_step = self.average_total_reward / steps_per_episode
		return self.episode_number, self.total_steps, reward_per_step, self.average_total_reward



from __future__ import absolute_import
import tensorflow as tf
from skopos.network.layers import Layer
from skopos.utils.preprocessing import Preprocessing

class SimpleLayer(Layer):
	"""docstring for SimpleLayer"""
	def __init__(self):
		super(SimpleLayer, self).__init__()
		
class FullyConnected(SimpleLayer):
	"""docstring for Fully Connected"""
	def __init__(self, size=100, weight_decay=None, initializer=None, initialize_bias=False):
		super(FullyConnected, self).__init__()
		self.size = size
		self.weight_decay = weight_decay
		self.initializer = initializer
		self.initialize_bias = initialize_bias

	def apply_layer(self, x):
		""" One fully connected layer """
		input_dim = x.get_shape()[1]
		fc_w = Layer.get_variables(name='weights', 
	    	shape=[input_dim, self.size], 
	    	weight_decay=self.weight_decay,
	    	initializer=self.initializer)
		if self.initialize_bias == True:
			fc_b = Layer.get_variables(name='bias', 
				shape=[self.size],
				weight_decay=self.weight_decay,
				initializer=self.initializer)
		else: 
			fc_b = Layer.get_variables(name='bias', 
				shape=[self.size],
				weight_decay=self.weight_decay)		
		fc_h = tf.matmul(x, fc_w) + fc_b
		return fc_h

	def get_scope(self):
		return "fc"
		
class Relu(SimpleLayer):
	"""docstring for Relu"""
	def __init__(self):
		super(Relu, self).__init__()

	def apply_layer(self, x):
		return tf.nn.relu(x)

	def get_scope(self):
		return "relu"

class Dropout(SimpleLayer):
	"""docstring for Relu"""
	def __init__(self, keep_probability=0.8):
		super(Dropout, self).__init__()
		self.keep_prob = keep_probability

	def apply_layer(self, x):
		return tf.nn.dropout(x, self.keep_prob)

	def get_scope(self):
		return "dropout"

class Sigmoid(SimpleLayer):
	"""docstring for Relu"""
	def __init__(self):
		super(Sigmoid, self).__init__()

	def apply_layer(self, x):
		return tf.sigmoid(x)

	def get_scope(self):
		return "sigmoid"

class Tanh(SimpleLayer):
	"""docstring for Relu"""
	def __init__(self):
		super(Tanh, self).__init__()

	def apply_layer(self, x):
		return tf.tanh(x)

	def get_scope(self):
		return "tanh"

		


from __future__ import absolute_import

import tensorflow as tf
import numpy as np

from skopos.memories.memory import ExperienceReplay, Episode
from skopos.utils.visualize import Plot
from skopos.agent.exploration_strategy import EGreedy
from skopos.agent.policies import NetworkBasedPolicy
from skopos.agent.environment_preprocessing import Environment

class Agent(object):
	"""docstring for Agent"""
	def __init__(self):
		super(Agent, self).__init__()

	def get_sess(self):
		return self.sess

	def get_coord(self):
		return self.coord

	def get_environment(self):
		return self.env

	def get_network(self):
		return self.network

	def get_processors(self):
		return self.processors

	def get_number_of_episodes(self):
		return self.number_of_episodes

	def get_max_episode_duration(self):
		return self.max_episode_duration

	def get_discount_factor(self):
		return self.discount_factor

	def get_pretrain_steps(self):
		return self.pretrain_steps

	def get_update_frequency(self):
		return self.update_frequency

	def get_batch_size(self):
		return self.batch_size

	def get_sequences(self):
		return self.sequences

	def get_model_folder(self):
		return self.model_folder

	def get_learner(self):
		return self.learner

	def get_memory(self):
		return self.memory

	def get_exploration_strategy(self):
		return self.exploration_strategy

	def get_policy(self):
		return self.policy

	def get_training_info(self):
		return self.training_info

	def get_tensorboard_visualization(self):
		return self.tensorboard_visualization

	def get_save_model(self):
		return self.save_model

	def get_restore_model(self):
		return self.restore_model

	def get_model_path(self):
		return self.model_path

	def get_save_model_frequency(self):
		return self.save_model_frequency

	def get_log_dir(self):
		return self.log_dir

	def restore_model_from_ckpt(self):
		if self.restore_model == True:
			self.saver = tf.train.import_meta_graph(self.model_path + '.meta')
			self.saver.restore(self.sess, tf.train.latest_checkpoint(self.model_folder))
			print("Model restored from checkpoint")
		return

	def save_model_ckpt(self):
		model_path = self.saver.save(self.sess, self.model_path)
		print("Model saved in: ", model_path)
		return



		
		

		




from __future__ import absolute_import

import tensorflow as tf
import numpy as np

class DoubleNetwork(object):
	"""docstring for DoubleNetwork"""
	def __init__(self, tau=5):
		super(DoubleNetwork, self).__init__()
		self.tau = tau
		self.total_updates = 0

	def set_network(self, network):
		self.network = network

	def get_network(self):
		return self.network

	def increment_total_updates(self):
		self.total_updates += 1

	def get_total_updates(self):
		return self.total_updates



from __future__ import absolute_import

import numpy as np

class Environment(object):
	"""docstring for EnvironmentPreprocessing"""
	def __init__(self, env):
		super(Environment, self).__init__()
		self.env = env

	def process_state(self, state):
		try:
			processed_state = np.zeros((1, self.env.observation_space.n))
			processed_state[0][state] = 1
		except AttributeError:
			processed_state = np.reshape(state, (-1, np.asarray(state).flatten().shape[0]))
		return processed_state
		


from __future__ import absolute_import

import tensorflow as tf
import numpy as np

from skopos.network.simple_layers import FullyConnected
from skopos.utils.preprocessing import Preprocessing
from skopos.techniques.learners import PolicyGradientLearner

class VanillaPolicyGradient(PolicyGradientLearner):
	"""docstring for VanillaPolicyGradient"""
	def __init__(self):
		super(VanillaPolicyGradient, self).__init__()

	def define_placeholders(self):
		x = tf.placeholder(np.float32, shape=[None, self.network.get_state_dimension()])
		y_ = tf.placeholder(np.float32, shape=[None])
		actions = tf.placeholder(shape=[None], dtype=tf.int32)
		return x, y_, actions

	def output(self, x, reuse):
		output_layer = FullyConnected(size=self.network.get_action_number())
		with tf.variable_scope(self.network.get_scope() + 'output_layer', reuse=reuse):
			out = output_layer.apply_layer(x)
		return tf.nn.softmax(out)

	def error_function(self, y, y_, actions):
		actions_ohe = tf.one_hot(actions, self.network.get_action_number(), dtype=tf.float32)
		responsible_outputs = tf.reduce_sum(y*actions_ohe, reduction_indices=[1])
		loss = - tf.reduce_sum(tf.log(responsible_outputs)*y_)
		return loss

	def prediction(self, out):
		return np.argmax(out, 1)

	def train(self, batch_size, number_of_sequences):
		batch = self.agent.get_memory().get_batches(batch_size, number_of_sequences)
		end_states = (1 - np.asarray(batch.get_dones()).astype(int))
		discounted_rewards = self.discount(batch.get_rewards(), self.agent.get_discount_factor()) * end_states

		starting_states = np.asarray(batch.get_starting_states()).reshape(-1, self.network.get_state_dimension())
		
		grads = self.agent.get_sess().run(self.network.optimizer, feed_dict={
			self.network.y_: discounted_rewards,
			self.network.actions: batch.get_actions(),
			self.network.x: starting_states})
		return


from distutils.core import setup
from setuptools import setup, find_packages

def check_dependencies():
    install_requires = []

    # Just make sure dependencies exist, I haven't rigorously
    # tested what the minimal versions that will work are
    # (help on that would be awesome)
    try:
        import numpy
    except ImportError:
        install_requires.append('numpy')
    try:
        import scipy
    except ImportError:
        install_requires.append('scipy')
    try:
        import pyqt
    except ImportError:
        install_requires.append('pyqt')
    try:
        import qt
    except ImportError:
        install_requires.append('qt')
    try:
        import mkl
    except ImportError:
        install_requires.append('mkl')
    try:
        import libpng
    except ImportError:
        install_requires.append('libpng')
    try:
        import icu
    except ImportError:
        install_requires.append('icu')
    try:
        import functools32
    except ImportError:
        install_requires.append('functools32')
    try:
        import freetype
    except ImportError:
        install_requires.append('freetype')
    try:
        import matplotlib
    except ImportError:
        install_requires.append('matplotlib')
    try:
        import tensorflow
    except ImportError:
        install_requires.append('tensorflow')
    return install_requires

install_requires = check_dependencies()

setup(
  name = 'skopos',
  # packages = ['skopos'],
  version = '0.4',
  description = 'Deep Reinforcement Learning Library',
  author = 'Skopos-team',
  packages=find_packages(),
  # packages=[package for package in find_packages() if package.startswith('skopos')],
  install_requires=install_requires,
  author_email = 'skopos.library@gmail.com',
  url = 'https://github.com/Skopos-team/Skopos', 
  license='Apache2',
  download_url = 'https://github.com/Skopos-team/Skopos/archive/0.4.tar.gz',
  keywords = ['testing', 'logging', 'example'],
  classifiers = ['Programming Language :: Python :: 2.7',
                  'Operating System :: POSIX',
                  'Operating System :: Unix',
                  'Operating System :: MacOS'
                  ],
)


from __future__ import absolute_import

import tensorflow as tf
import numpy as np
import random

class Policy(object):
	"""docstring for Policy"""
	def __init__(self):
		super(Policy, self).__init__()

	def get_environment(self):
		return self.env

	def set_env(self, env):
		self.env = env

	def set_network(self, network):
		self.network = network

	def set_session(self, sess):
		self.sess = sess

class RandomPolicy(Policy):
	"""docstring for RandomPolicy"""
	def __init__(self):
		super(RandomPolicy, self).__init__()

	def evaluate(self, state):
		action = random.randint(0, self.env.action_space.n - 1)
		return action, np.ones(self.env.action_space.n)

class NetworkBasedPolicy(Policy):
	"""docstring for NetworkBasedPolicy"""
	def __init__(self):
		super(NetworkBasedPolicy, self).__init__()
		
	def evaluate(self, state):
		actions = self.sess.run(self.network.out, feed_dict={self.network.x : state})
		return int(self.network.get_learner().prediction(actions)), actions
		